(* 
 * Copyright (c) 2000 Carnegie Mellon University.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The name "Carnegie Mellon University" must not be used to
 *    endorse or promote products derived from this software without
 *    prior written permission. For permission or any other legal
 *    details, please contact  
 *	Office of Technology Transfer
 *	Carnegie Mellon University
 *	5000 Forbes Avenue
 *	Pittsburgh, PA  15213-3890
 *	(412) 268-4387, fax: (412) 268-7395
 *	tech-transfer@andrew.cmu.edu
 *
 * 4. Redistributions of any form whatsoever must retain the following
 *    acknowledgment:
 *    "This product includes software developed by Computing Services
 *     at Carnegie Mellon University (http://www.cmu.edu/computing/)."
 *
 * CARNEGIE MELLON UNIVERSITY DISCLAIMS ALL WARRANTIES WITH REGARD TO
 * THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS, IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY BE LIABLE
 * FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN
 * AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING
 * OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *)

structure AcapTime :> ACAPTIME =
struct
  (* think about using type acaptime = string ? *)
  type acaptime = Time.time * int

  exception BadTimeString

  open Date

  val old = (Time.zeroTime, 0 : int)

  val future = (Time.fromSeconds (valOf Int32.maxInt), 0 : int)

  val tick = Time.fromSeconds 15

  val reqCh = (CML.channel ()) : acaptime SyncVar.ivar CML.chan

  val round = Time.fromSeconds o Time.toSeconds
  val now = round o Time.now

  fun server (time as (cur, t), timeout) = 
    let
      val respCh = CML.channel ()

      val tevt = CML.wrap(CML.atTimeEvt timeout, 
			  fn () => let val new = now () 
				   in server ((new, 0), Time.+(new, tick)) end)

      val revt = CML.wrap(CML.recvEvt reqCh, 
			  fn (replyV) => (SyncVar.iPut (replyV, time);
					  server ((cur, t+1), timeout)))
    in
      CML.select [tevt, revt]
    end

  fun init () = 
    let
      val new = now ()
    in ignore (CML.spawnc server ((new, 0), Time.+(new, tick))) end

  fun now () = 
    let
      val replyV = SyncVar.iVar ()
    in
      CML.send (reqCh, replyV);
      SyncVar.iGet replyV
    end

  fun compare ((t1, c1), (t2, c2)) = 
    case Time.compare (t1, t2) of
      EQUAL => Int.compare (c1, c2)
    | other => other

  fun intToString p = (StringCvt.padLeft #"0" p) o Int.toString
  fun int32ToString p = (StringCvt.padLeft #"0" p) o Int32.toString

  fun monthToString Jan = "01"
    | monthToString Feb = "02"
    | monthToString Mar = "03"
    | monthToString Apr = "04"
    | monthToString May = "05"
    | monthToString Jun = "06"
    | monthToString Jul = "07"
    | monthToString Aug = "08"
    | monthToString Sep = "09"
    | monthToString Oct = "10"
    | monthToString Nov = "11"
    | monthToString Dec = "12"

  val iTS4 = intToString 4
  val iTS2 = intToString 2
  val i32TS6 = intToString 6

  fun toString (t, count) = (* sigh *)
    let
      val d = Date.fromTimeUniv t
    in
      String.concat [iTS4 (year d),
		     monthToString (month d),
		     iTS2 (day d),
		     iTS2 (hour d),
		     iTS2 (minute d),
		     iTS2 (second d),
		     i32TS6 count]
    end

  exception BadMonth

  fun fromString "00000101000000" = old
    | fromString s =
    let
      fun splitString n s = (String.extract (s, 0, SOME n),
			     String.extract (s, n, NONE))
      
      fun convMonth "01" = Jan
	| convMonth "02" = Feb
	| convMonth "03" = Mar
	| convMonth "04" = Apr
	| convMonth "05" = May
	| convMonth "06" = Jun
	| convMonth "07" = Jul
	| convMonth "08" = Aug
	| convMonth "09" = Sep
	| convMonth "10" = Oct
	| convMonth "11" = Nov
	| convMonth "12" = Dec
	| convMonth _ = raise BadMonth

      val (year, s) = splitString 4 s
      val (month, s) = splitString 2 s
      val (day, s) = splitString 2 s
      val (hour, s) = splitString 2 s
      val (minute, s) = splitString 2 s
      val (second, smallt) = splitString 2 s

      val conv = (valOf o Int.fromString) 
      val year = conv year
      val month = convMonth month
      val day = conv day
      val hour = conv hour
      val minute = conv minute
      val second = conv second
      val smallt = getOpt(Int.fromString smallt, 0)

      val year = if year < 1970 then 1970 else year
      val date = Date.date {year=year, month=month, day=day, hour=hour,
			    minute=minute, second=second, 
			    offset=(SOME Time.zeroTime)}

      val time = (Date.toTime date, smallt)
    in
      time
    end handle _ => raise BadTimeString
    
  fun toValue t = Value.Single (toString t)
end
